#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""
Module contains definitions for BLE standard HCI commands.
"""
# pylint: disable=too-many-arguments
from typing import List, Optional, Tuple, Union

from ._hci_logger import get_formatted_logger
from ._transport import SerialUartTransport
from .constants import PayloadOption, PhyOption
from .data_params import AdvParams, ConnParams, ScanParams
from .hci_packets import CommandPacket, EventPacket
from .packet_codes import StatusCode
from .packet_defs import OCF, OGF
from .utils import to_le_nbyte_list


class BleStandardCmds:
    """Definitions for BLE standard HCI commands.

    Class contains functions used to implement BLE standard HCI
    commands. Used as a parent for the full Analog Devices BLE
    HCI class.

    Arguments
    ---------
    port : SerialUartTransport
        Serial port interfacing object.
    logger_name: str
        Name used to reference the HCI logger.

    Attributes
    ----------
    port : SerialUartTransport
        Serial port interfacing object.
    logger : logging.Logger
        HCI logging object referenced by the `name` argument.

    """

    def __init__(self, port: SerialUartTransport, logger_name: str):
        self.port = port
        self.logger = get_formatted_logger(name=logger_name)

    def send_le_controller_command(
        self, ocf: OCF, params: List[int] = None, return_evt: bool = False
    ) -> Union[StatusCode, EventPacket]:
        """Send an LE Controller command to the test board.

        Sends a command from the OGF LE Controller subgroup to the DUT.

        Parameters
        ----------
        ocf : OCF
            Opcode command field value for the desired HCI command.
        params : List[int], optional
            Command parameters as single-byte values.
        return_evt : bool, optional
            If true, function returns full `EventPacket` object. If
            false, function returns only the status code.

        Returns
        -------
        Union[StatusCode, EventPacket]
            If `return_evt` argument is true, the full return packet
            from the DUT. If `return_evt` argument is false, the return
            packet status code.


        """
        cmd = CommandPacket(OGF.LE_CONTROLLER, ocf, params=params)
        if return_evt:
            return self.port.send_command(cmd)

        return self.port.send_command(cmd).status

    def send_link_control_command(
        self, ocf: OCF, params: List[int] = None, return_evt: bool = False
    ) -> Union[StatusCode, EventPacket]:
        """Send a Link Control command to the test board.

        Sends a command from the OGF Link Control subgroup to the DUT.

        Parameters
        ----------
        ocf : OCF
            Opcode command field value for the desired HCI command.
        params : List[int], optional
            Command parameters as single-byte values.
        return_evt : bool, optional
            If true, function returns full `EventPacket` object. If
            false, function returns only the status code.

        Returns
        -------
        Union[StatusCode, EventPacket]
            If `return_evt` argument is true, the full return packet
            from the DUT. If `return_evt` argument is false, the return
            packet status code.


        """
        cmd = CommandPacket(OGF.LINK_CONTROL, ocf, params=params)
        if return_evt:
            return self.port.send_command(cmd)

        return self.port.send_command(cmd).status

    def send_controller_command(
        self, ocf: OCF, params: List[int] = None, return_evt: bool = False
    ) -> Union[StatusCode, EventPacket]:
        """Send a Controller command to the test board.

        Sends a command from the OGF Controller subgroup to the DUT.

        Parameters
        ----------
        ocf : OCF
            Opcode command field value for the desired HCI command.
        params : List[int], optional
            Command parameters as single-byte values.
        return_evt : bool, optional
            If true, function returns full `EventPacket` object. If
            false, function returns only the status code.

        Returns
        -------
        Union[StatusCode, EventPacket]
            If `return_evt` argument is true, the full return packet
            from the DUT. If `return_evt` argument is false, the return
            packet status code.


        """
        cmd = CommandPacket(OGF.CONTROLLER, ocf, params=params)
        if return_evt:
            return self.port.send_command(cmd)

        return self.port.send_command(cmd).status

    def set_adv_params(self, adv_params: AdvParams = AdvParams()) -> StatusCode:
        """Set test board advertising parameters.

        Sends a command to the DUT, telling it to set the advertising
        parameters to the given values.

        Parameters
        ----------
        adv_params : AdvParams, optional
            Dataclass object containing the desired advertising
            parameters.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        params = to_le_nbyte_list(adv_params.interval_min, 2)
        params.extend(to_le_nbyte_list(adv_params.interval_max, 2))
        params.extend(
            [
                adv_params.adv_type,
                adv_params.own_addr_type.value,
                adv_params.peer_addr_type.value,
            ]
        )
        params.extend(to_le_nbyte_list(adv_params.peer_addr, 6))
        params.extend([adv_params.channel_map, adv_params.filter_policy])

        return self.send_le_controller_command(
            OCF.LE_CONTROLLER.SET_ADV_PARAM, params=params
        )

    def enable_adv(self, enable: bool) -> StatusCode:
        """Command board to start/stop advertising.

        Sends a command to the DUT, telling it to either start or
        stop advertising based on the the `enable` argument.

        Parameters
        ----------
        enable : bool
            Enable advertising?

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        return self.send_le_controller_command(
            OCF.LE_CONTROLLER.SET_ADV_ENABLE, params=int(enable)
        )

    def set_scan_params(self, scan_params: ScanParams = ScanParams()) -> StatusCode:
        """Set test board scanning parameters.

        Sends a command to the DUT, telling it to set the scanning
        parameters to the given values.

        Parameters
        ----------
        scan_params : ScanParams, optional
            Dataclass object containing the desired scanning parameters.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        params = [scan_params.scan_type]
        params.extend(to_le_nbyte_list(scan_params.scan_interval, 2))
        params.extend(to_le_nbyte_list(scan_params.scan_window, 2))
        params.append(scan_params.addr_type.value)
        params.append(scan_params.filter_policy)

        return self.send_le_controller_command(
            OCF.LE_CONTROLLER.SET_SCAN_PARAM, params=params
        )

    def enable_scanning(
        self, enable: bool, filter_duplicates: bool = False
    ) -> StatusCode:
        """Command board to start/stop scanning.

        Sends a command to the DUT, telling it to either start or
        stop scanning based on the `enable` argument.

        Parameters
        ----------
        enable : bool
            Enable scanning?
        filter_duplicates : bool, optional
            Filter duplicates?

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        params = [int(enable), int(filter_duplicates)]
        return self.send_le_controller_command(
            OCF.LE_CONTROLLER.SET_SCAN_ENABLE, params=params
        )

    def create_connection(
        self, conn_params: ConnParams = ConnParams(0x0)
    ) -> StatusCode:
        """Command board to connect with a peer device.

        Sends a command to the DUT, telling it to create a connection
        to a peer device based on the given connection parameters.

        Parameters
        ----------
        conn_params : ConnParams, optional
            Dataclass object containing the desired connection parameters.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        params = to_le_nbyte_list(conn_params.scan_interval, 2)
        params.extend(to_le_nbyte_list(conn_params.scan_window, 2))
        params.append(conn_params.init_filter_policy)
        params.append(conn_params.peer_addr_type.value)
        params.extend(to_le_nbyte_list(conn_params.peer_addr, 6))
        params.append(conn_params.own_addr_type.value)
        params.extend(to_le_nbyte_list(conn_params.conn_interval_min, 2))
        params.extend(to_le_nbyte_list(conn_params.conn_interval_max, 2))
        params.extend(to_le_nbyte_list(conn_params.max_latency, 2))
        params.extend(to_le_nbyte_list(conn_params.sup_timeout, 2))
        params.extend(to_le_nbyte_list(conn_params.min_ce_length, 2))
        params.extend(to_le_nbyte_list(conn_params.max_ce_length, 2))

        return self.send_le_controller_command(
            OCF.LE_CONTROLLER.CREATE_CONN, params=params
        )

    def set_default_phy(
        self, all_phys: int = 0x0, tx_phys: int = 0x7, rx_phys: int = 0x7
    ) -> StatusCode:
        """Set defaults for ALL, TX, and RX PHYs.

        Sends a command to the DUT, telling it to set the default behavior
        for ALL, TX, and RX PHYs in accordance with the given values.

        Parameters
        ----------
        all_phys : int, optional
            Value describing desired behavior of all PHYs.
        tx_phys : int, optional
            Value describing desired behavior of TX PHYs.
        rx_phys : int, optional
            Value describing desired behavior of RX PHYs.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        params = [all_phys, tx_phys, rx_phys]
        return self.send_le_controller_command(
            OCF.LE_CONTROLLER.SET_DEF_PHY, params=params
        )

    def set_data_len(
        self, handle: int = 0x0000, tx_octets: int = 0xFB00, tx_time: int = 0x9042
    ) -> StatusCode:
        """Set the maximum TX payload size and transmit time.

        Sends a command to the DUT, telling it to set the maximum TX
        payload size and transmit time to the given values.

        Parameters
        ----------
        handle : int, optional
            Connection handle.
        tx_octets : int, optional
            Desired maximum number of payload octets.
        tx_time : int, optional
            Desired maximum TX time.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        params = to_le_nbyte_list(handle, 2)
        params.extend(to_le_nbyte_list(tx_octets, 2))
        params.extend(to_le_nbyte_list(tx_time, 2))
        return self.send_le_controller_command(
            OCF.LE_CONTROLLER.SET_DATA_LEN, params=params
        )

    def set_phy(
        self,
        handle: int = 0x0000,
        all_phys: int = 0x0,
        tx_phys: int = 0x7,
        rx_phys: int = 0x7,
        phy_opts: int = 0x0,
    ) -> StatusCode:
        """Set the PHY preferences for a connection.

        Sends a command to the DUT, telling it to set the PHY preferences
        for the indicated connection in accordance with the given values.

        Parameters
        ----------
        handle : int, optional
            The handle to the desired connection.
        all_phys : int, optional
            Behavior settings for all PHYs. Indicates if a PHY preference
            exists for both RX and TX PHYs.
        tx_phys : PhyOption, optional
            PHY preference for TX PHYs.
        rx_phys : PhyOption, optional
            PHY preference for RX PHYs.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        if not 0 <= all_phys <= 3:
            self.logger.warning(
                "Invalid all PHYs option (%i), must be between 0x0 and 0x3. Defaulting to 0x0.",
                all_phys,
            )
        if all_phys in (1, 3) and not 0 < tx_phys <= 7:
            self.logger.warning(
                "Invalid TX PHY option (%i), must be between 0x1 and 0x7. Defaulting to 0x7.",
                tx_phys,
            )
            tx_phys = 0x7
        if all_phys in (2, 3) and not 0 < rx_phys <= 7:
            self.logger.warning(
                "Invalid RX PHY option (%i), must be between 0x1 and 0x7. Defaulting to 0x7.",
                rx_phys,
            )
            rx_phys = 0x7
        if phy_opts not in (0, 1, 2, 4):
            self.logger.warning(
                "Invalid PHY options (%i), must be 0x0, 0x1, 0x2, or 0x4. Defaulting to 0x0.",
                phy_opts,
            )
            phy_opts = 0x0

        params = to_le_nbyte_list(handle, 2)
        params.extend([all_phys, tx_phys, rx_phys])
        params.extend(to_le_nbyte_list(phy_opts, 2))

        return self.send_le_controller_command(OCF.LE_CONTROLLER.SET_PHY, params=params)

    def tx_test(
        self,
        channel: int = 0,
        phy: Union[PhyOption, int] = PhyOption.PHY_1M,
        payload: Union[PayloadOption, int] = PayloadOption.PLD_PRBS9,
        packet_len: int = 0,
    ) -> StatusCode:
        """Start a transmitter test.

        Sends a command to the DUT, telling it to start a DTM transmitter
        test in accordance with the given parameters.

        Parameters
        ----------
        channel : int, optional
            The channel on which transmission should take place.
        phy : Union[PhyOption,int], optional
            The PHY that should be used by the transmitter.
        payload : PayloadOption, optional
            The packet payload type that should be used.
        packet_len : int, optional
            The desired length of the transmitted packets.

        Returns
        -------
        StatusCode
            The return packet status code.

        """

        if isinstance(payload, PayloadOption):
            payload = payload.value
        if isinstance(phy, PhyOption):
            phy = phy.value

        params = [channel, packet_len, payload, phy]
        return self.send_le_controller_command(
            OCF.LE_CONTROLLER.ENHANCED_TRANSMITTER_TEST, params=params
        )

    def rx_test(
        self,
        channel: int = 0,
        phy: Union[PhyOption, int] = PhyOption.PHY_1M,
        modulation_idx: int = 0,
    ) -> StatusCode:
        """Start a receiver test.

        Sends a command to the DUT, telling it to start a DTM receiver
        test in accordance with the given parameters.

        Parameters
        ----------
        channel : int, optional
            The channel on which the receiver should listen for packets.
        phy : Union[PhyOption,int], optional
            The PHY that should be used by the receiver.
        modulation_idx : float, optional
            The expected modulation index of the transmitter. Indicates
            whether the modulation index is standard (0) or stable (1).

        Returns
        -------
        StatusCode
            The return packet status code.

        """

        if phy == PhyOption.PHY_CODED_S2:
            phy = PhyOption.PHY_CODED_S8

        if isinstance(phy, PhyOption):
            phy = phy.value

        params = [channel, phy, modulation_idx]
        return self.send_le_controller_command(
            OCF.LE_CONTROLLER.ENHANCED_RECEIVER_TEST, params=params
        )

    def end_test(self) -> Tuple[int, StatusCode]:
        """End the current test.

        Sends a command to the DUT, telling it to end the current
        DTM test.

        Returns
        -------
        StatusCode
            The return packet status code.
        int
            The number of packets received correctly during the test. If
            ending a TX test, this value will be 0.

        """
        evt = self.send_le_controller_command(
            OCF.LE_CONTROLLER.TEST_END, return_evt=True
        )
        rx_ok = evt.get_return_params()

        return rx_ok, evt.status

    def disconnect(self, handle: int = 0x0000, reason: int = 0x16) -> StatusCode:
        """Disconnect from an existing connection.

        Sends a command to the DUT, telling it to disconnect from
        the indicated connection for the given reason.

        Parameters
        ----------
        handle : int, optional
            The handle to the desired connection.
        reason : int, optional
            The reason for the disconnection.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        params = to_le_nbyte_list(handle, 2)
        params.append(reason)
        return self.send_link_control_command(
            OCF.LINK_CONTROL.DISCONNECT, params=params
        )

    def reset(self) -> StatusCode:
        """Reset board controller/link layer.

        Sends a command to the DUT, telling it that the controller
        and the link layer should be reset. On-board implementation
        may vary, meaning this command does not necessarily perform
        a full hardware reset.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        return self.send_controller_command(OCF.CONTROLLER.RESET)

    def set_event_mask(
        self, mask: int, mask_pg2: Optional[int] = None
    ) -> Union[StatusCode, Tuple[StatusCode, StatusCode]]:
        """Enable/disable events the board can generate.

        Sends a command to the DUT, telling it to enable/disable
        events that can be generated and returned to the host in
        accordance with the given mask. If a page2 mask if provided,
        then the command which sets the page2 masks will also be sent.

        Parameters
        ----------
        mask : int
            Mask indicating the desired events. Setting a bit to `1`
            enables the corresponding event. Setting the bit to `0`
            disables it.
        mask_pg2 : Optional[int], optional
            Mask indicating the desired events for the second event
            mask page. Setting a bit to `1` enables the corresponding
            event. Setting the bit to `0` disables it.

        Returns
        -------
        Union[StatusCode, Tuple[StatusCode, StatusCode]]
            The return packet status codes(s). If both page1 and page2
            were set, the first return is the status code for the page1
            command and the second is the status code for the page2
            command.

        """
        params = to_le_nbyte_list(mask, 8)
        status = self.send_controller_command(
            OCF.CONTROLLER.SET_EVENT_MASK, params=params
        )

        if mask_pg2:
            params = to_le_nbyte_list(mask_pg2, 8)
            return (
                status,
                self.send_controller_command(
                    OCF.CONTROLLER.SET_EVENT_MASK_PAGE2, params=params
                ),
            )

        return status

    def set_event_mask_le(self, mask: int) -> StatusCode:
        """Enable/disable LE events the board can generate.

        Sends a command to the DUT, telling it to enable/disable
        LE events that can be generated and returned to the host
        in accordance with the given mask.

        Parameters
        ----------
        mask : int
            Mask indicating the desired LE events. Setting a bit
            to `1` enables the corresponding event. Setting the
            bit to `0` disables it.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        params = to_le_nbyte_list(mask, 8)
        return self.send_le_controller_command(
            OCF.LE_CONTROLLER.SET_EVENT_MASK, params=params
        )



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""
connection.py

Description: Simple example showing creation of a connection and getting packet error metrics

"""

import time

from ble_hci import BleHci

# Switch out for serial port used to connect over HCI
MASTER_HCI_PORT = "/dev/serial/by-id/usb-FTDI_FT230X_Basic_UART_D3073IDU-if00-port0"
SLAVE_HCI_PORT = "/dev/serial/by-id/usb-FTDI_FT231X_USB_UART_D30AKVUY-if00-port0"


def evt_callback(data):
    print(data)


def main():
    """MAIN"""
    master = BleHci(MASTER_HCI_PORT, evt_callback=evt_callback)
    slave = BleHci(SLAVE_HCI_PORT, evt_callback=evt_callback)

    master.reset()
    slave.reset()

    master_addr = 0x001234887733
    slave_addr = 0x111234887733

    master.set_adv_tx_power(-10)
    slave.set_adv_tx_power(-10)

    slave.set_address(slave_addr)
    master.set_address(master_addr)

    slave.start_advertising(connect=True)
    master.init_connection(addr=slave_addr)

    while True:
        slave_stats, _ = slave.get_conn_stats()
        master_stats, _ = master.get_conn_stats()

        if slave_stats.rx_data and master_stats.rx_data:
            print(f"Slave PER {slave_stats.per()}")
            print(f"Master PER {master_stats.per()}")
            break

        time.sleep(0.5)

    master.disconnect()
    slave.disconnect()

    master.reset()
    slave.reset()


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""Contains data classes used for HCI function parameters/returns."""
from dataclasses import dataclass

# pylint: disable=too-many-arguments,too-many-locals,too-many-instance-attributes
from typing import Optional

from .constants import AddrType


@dataclass
class AdvParams:
    """Advertising parameters data container."""

    interval_min: int = 0x60
    """Minimum advertising interval."""

    interval_max: int = 0x60
    """Maximum advertising interval."""

    adv_type: int = 0x3
    """Advertising type."""

    own_addr_type: AddrType = AddrType.PUBLIC
    """Own device address type."""

    peer_addr_type: AddrType = AddrType.PUBLIC
    """Connectable peer device address type."""

    peer_addr: int = 0
    """Connectable peer device address."""

    channel_map: int = 0x7
    """Advertising channel map."""

    filter_policy: int = 0
    """Advertising filter policy."""

    def __repr__(self) -> str:
        print_lns = []
        for key, val in self.__dict__.items():
            if val is None:
                continue
            print_lns.append(f"{key}:  {val}")

        return "\n".join(print_lns)


@dataclass
class ScanParams:
    """Scan parameters data container."""

    scan_type: int = 0x1
    """Scan type."""

    scan_interval: int = 0x100
    """Scan interval."""

    scan_window: int = 0x100
    """Scan duration."""

    addr_type: AddrType = AddrType.PUBLIC
    """Own device address type."""

    filter_policy: int = 0x0
    """Scanning filter policy."""

    def __repr__(self) -> str:
        print_lns = []
        for key, val in self.__dict__.items():
            if val is None:
                continue
            print_lns.append(f"{key}:  {val}")

        return "\n".join(print_lns)


@dataclass
class ConnParams:
    """Connection parameters data container."""

    peer_addr: int
    """Connectable peer device address."""

    scan_interval: int = 0x100
    """Scan interval."""

    scan_window: int = 0x100
    """Scan duration."""

    init_filter_policy: int = 0x0
    """Initiator filter policy."""

    peer_addr_type: AddrType = AddrType.PUBLIC
    """Connectable peer device address type."""

    own_addr_type: AddrType = AddrType.PUBLIC
    """Own device address type."""

    conn_interval_min: int = 0x6
    """Connection interval minimum."""

    conn_interval_max: int = 0x6
    """Connection interval maximum."""

    max_latency: int = 0x0000
    """Maximum peripheral latency."""

    sup_timeout: int = 0x64
    """Supervision timeout."""

    min_ce_length: int = 0x0F10
    """Minimum connection event length."""

    max_ce_length: int = 0x0F10
    """Maximum connection event length."""

    def __post_init__(self):
        if not 0x4 <= self.scan_interval <= 0x4000:
            raise ValueError("Scane interval must be between 0x4 - 0x4000")
        if not 0x4 <= self.scan_window <= 0x4000:
            raise ValueError("Scane window must be between 0x4 - 0x4000")

        if not self.init_filter_policy in [0, 1]:
            raise ValueError(
                f"Init filter policy must be 0x0 or 0x1 {self.init_filter_policy}"
            )

        if not isinstance(self.peer_addr_type, AddrType):
            raise TypeError(
                "Attribute peer_addr_type must be of type AddrType, ",
                f"not {type(self.peer_addr_type).__name__}",
            )

        if self.peer_addr > 2**48 - 1:
            raise ValueError("Peer address must be representable by 6 octets")

        if not isinstance(self.own_addr_type, AddrType):
            raise TypeError(
                "Attribute own_addr_type must be of type AddrType, ",
                f"not {type(self.own_addr_type).__name__}",
            )

        if not 0x6 <= self.conn_interval_max <= 0xC80:
            raise ValueError("Connection interval min must be between 0x6 - 0xC80")

    def __repr__(self) -> str:
        print_lns = []
        for key, val in self.__dict__.items():
            if val is None:
                continue
            print_lns.append(f"{key}:  {val}")

        return "\n".join(print_lns)


class DataPktStats:
    """Generic data stats container for CM and DTM."""

    def __init__(
        self,
        rx_data: int,
        rx_data_crc: int,
        rx_data_timeout: int,
        tx_data: int,
        err_data: int,
        rx_setup: int,
        tx_setup: int,
        rx_isr: int,
        tx_isr: int,
    ) -> None:
        self.rx_data = rx_data
        """Number of packets received correctly."""

        self.rx_data_crc = rx_data_crc
        """Number of packets received with a CRC error."""

        self.rx_data_timeout = rx_data_timeout
        """Number of RX timeouts."""

        self.tx_data = tx_data
        """Number of correctly transmitted packets."""

        self.err_data = err_data
        """Number of data transaction errors."""

        self.rx_setup = rx_setup
        """RX packet setup watermark in microseconds."""

        self.tx_setup = tx_setup
        """TX packet setup watermark in microseconds."""

        self.rx_isr = rx_isr
        """RX ISR processing watermark in microseconds."""

        self.tx_isr = tx_isr
        """TX ISR processing watermark in microseconds."""

    def __repr__(self) -> str:
        print_lns = []
        for key, val in self.__dict__.items():
            if val is None:
                continue
            print_lns.append(f"{key}:  {val}")

        return "\n".join(print_lns)

    def per(self, peer_tx_data: Optional[int] = None) -> float:
        """Calculate PER.

        Calculates the Packet Error Rate of the current set of
        statistics. If number of peer TX packets is not provided,
        the value is inferred from the number of correctly received
        packets, the number of CRC errors, and the number of timeouts
        recorded by the receiver.

        Parameters
        ----------
        peer_tx_data : Optional[int], optional
            Number of packets transmitted by the peer device.

        Returns
        -------
        float
            Calculated PER value.

        """
        if peer_tx_data:
            return 100 - 100 * (self.rx_data / peer_tx_data)
        return 100 - self.rx_data / (
            self.rx_data + self.rx_data_crc + self.rx_data_timeout
        )


@dataclass
class AdvPktStats:
    """Advertising statistics data container."""

    tx_adv: int = None
    """Number of sent advertising packets."""

    rx_req: int = None
    """Number of advertising requests received correctly."""

    rx_req_crc: int = None
    """Number of advertising requests received with a CRC error."""

    rx_req_timeout: int = None
    """Number of RX timeouts."""

    tx_resp: int = None
    """Number of response packets sent."""

    err_adv: int = None
    """Number of advertising transaction errors."""

    rx_setup: int = None
    """RX packet setup watermark in microseconds."""

    tx_setup: int = None
    """TX packet setup watermark in microseconds."""

    rx_isr: int = None
    """RX ISR processing watermark in microseconds."""

    tx_isr: int = None
    """TX ISR processing watermark in microseconds."""

    tx_chain: Optional[int] = None
    """
    Number of chain packets sent.
    
    .. note::
        Value is only returned when retrieving auxiliary advertising statistics.
    
    """

    def __repr__(self) -> str:
        print_lns = []
        for key, val in self.__dict__.items():
            if val is None:
                continue
            print_lns.append(f"{key}:  {val}")

        return "\n".join(print_lns)


@dataclass
class ScanPktStats:
    """Scanning statistics data container."""

    rx_adv: int = None
    """Number of advertising packets received correctly."""

    rx_adv_crc: int = None
    """Number of advertising packets received with a CRC error."""

    rx_adv_timeout: int = None
    """Number of RX timeouts."""
    tx_req: int = None
    """Number of advertising requests sent."""
    rx_rsp: int = None
    """Number of advertising response packets received correctly."""

    rx_rsp_crc: int = None
    """Number of advertising response packets received with a CRC error."""

    rx_rsp_timeout: int = None
    """Number of advertising response RX timeouts."""
    err_scan: int = None
    """Number of scan transaction errors."""
    rx_setup: int = None
    """RX packet setup watermark in microseconds."""

    tx_setup: int = None
    """TX packet setup watermark in microseconds."""
    rx_isr: int = None
    """RX ISR processing watermark in microseconds."""

    tx_isr: int = None
    """TX ISR processing watermark in microseconds."""
    rx_chain: Optional[int] = None
    """
    Number of chain packets received correctly.
    
    .. note::
        Value is only returned when retrieving auxiliary scanning statistics.
    
    """

    rx_chain_crc: Optional[int] = None
    """
    Number of chain packets received with a CRC error.
    
    .. note::
        Value is only returned when retrieving auxiliary scanning statistics.

    """

    rx_chain_timeout: Optional[int] = None
    """
    Number of chain packet RX timeouts.
    
    .. note::
        Value is only returned when retrieving auxiliary scanning statistics.
    
    """

    def __repr__(self) -> str:
        print_lns = []
        for key, val in self.__dict__.items():
            if val is None:
                continue
            print_lns.append(f"{key}:  {val}")

        return "\n".join(print_lns)


@dataclass
class MemPktStats:
    """Memory statistics data container."""

    stack: int = None
    """Number of bytes used by the stack."""

    sys_assert_cnt: int = None
    """Number of times assertions hit."""

    free_mem: int = None
    """Memory free for stack usage."""

    used_mem: int = None
    """Memory used by stack."""

    max_connections: int = None
    """Maximum number of connections allowed."""

    conn_ctx_size: int = None
    """Number of bytes used for connection context."""

    cs_watermark_lvl: int = None
    """Critical section watermark duration in microseconds."""

    ll_watermark_lvl: int = None
    """Link layer handler watermark level in microseconds."""

    sch_watermark_lvl: int = None
    """Scheduler handler watermark level in microseconds."""

    lhci_watermark_lvl: int = None
    """LHCI handler watermark level in microseconds."""

    max_adv_sets: int = None
    """Maximum number of advertising sets."""

    adv_set_ctx_size: int = None
    """Advertising set context size in bytes."""

    ext_scan_max: int = None
    """Maximum number of extended scanners."""

    ext_scan_ctx_size: int = None
    """Extended scanners context size in bytes."""

    ext_init_ctx_size: int = None
    """Extended initiators context size in bytes."""

    max_num_ext_init: int = None
    """Maximum number of extended initiators."""

    max_per_scanners: int = None
    """Maximum number of periodic scanners."""

    per_scan_ctx_size: int = None
    """Periodic scanners context size in bytes."""

    max_cig: int = None
    """Maximim number of CIG."""

    cig_ctx_size: int = None
    """CIG context size in bytes."""

    cis_ctx_size: int = None
    """CIS context size in bytes."""

    def __repr__(self) -> str:
        print_lns = []
        for key, val in self.__dict__.items():
            if val is None:
                continue
            print_lns.append(f"{key}:  {val}")

        return "\n".join(print_lns)


@dataclass
class PduPktStats:
    """PDU statistics data container."""

    fail_pdu: int = None
    """Number of PDUs failing PDU type filter."""

    pass_pdu: int = None
    """Number of PDUs passing PDU type filter."""

    fail_whitelist: int = None
    """Number of PDUs failing whitelist filter."""

    pass_whitelist: int = None
    """Number of PDUs passing whitelist filter."""

    fail_peer_addr_match: int = None
    """Number of PDUs failing peer address match."""

    pass_peer_addr_match: int = None
    """Number of PDUs passing peer address match."""

    fail_local_addr_match: int = None
    """Number of PDUs failing local address match."""

    pass_local_addr_match: int = None
    """Number of PDUs passing local address match."""

    fail_peer_rpa_verify: int = None
    """Number of peer RPAs failing verification."""

    pass_peer_rpa_verify: int = None
    """Number of peer RPAs passing verification."""

    fail_local_rpa_verify: int = None
    """Number of local RPAs failing verification."""

    pass_local_rpa_verify: int = None
    """Number of local RPAs passing verification."""

    fail_peer_priv_addr: int = None
    """Number of peer addresses failing RPA requirements."""

    fail_local_priv_addr: int = None
    """Number of local addresses failing RPA requirements."""

    fail_peer_addr_res_req: int = None
    """Number of PDUs failing required peer address resolution."""

    pass_peer_addr_res_req: int = None
    """Number of PDUs passing required peer address resolution."""

    pass_local_addr_res_opt: int = None
    """Number of PDUs passing optional local address resolution."""

    peer_res_addr_pend: int = None
    """Number of peer address resolutions pended."""

    local_res_addr_pend: int = None
    """Number of local address resolutions pended."""

    def __repr__(self) -> str:
        print_lns = []
        for key, val in self.__dict__.items():
            if val is None:
                continue
            print_lns.append(f"{key}:  {val}")

        return "\n".join(print_lns)


@dataclass
class TestReport:
    """ISO/ACL test report data container."""

    rx_pkt_count: int = None
    """Number of packets received."""

    rx_oct_count: int = None
    """Number of octets received."""

    gen_pkt_count: int = None
    """Number of generated packets."""

    gen_oct_count: int = None
    """Number of generated octets."""

    def __repr__(self) -> str:
        print_lns = []
        for key, val in self.__dict__.items():
            if val is None:
                continue
            print_lns.append(f"{key}:  {val}")

        return "\n".join(print_lns)


@dataclass
class PoolStats:
    """Memory pool statistics data container."""

    buf_size: int = None
    """Buffer size."""

    num_buf: int = None
    """Number of buffers."""

    num_alloc: int = None
    """Number of allocations."""

    max_alloc: int = None
    """Maximum number of allocations."""

    max_req_len: int = None
    """Maximum required length."""

    def __repr__(self) -> str:
        print_lns = []
        for key, val in self.__dict__.items():
            if val is None:
                continue
            print_lns.append(f"{key}:  {val}")

        return "\n".join(print_lns)



import secrets
import time
import unittest

from max_ble_hci import BleHci
from max_ble_hci import packet_codes as pc

from max_ble_hci.constants import PhyOption, PubKeyValidateMode

PORT = ""

hci1 = BleHci(PORT, id_tag="hci1")

MAX_U32 = 0xFFFFFFFF
A32 = 0xAAAAAAAA


class TestHci(unittest.TestCase):
    def test_reset(self):
        # Reset puts code into nice condition, make sure it works before any tests
        self.assertEqual(hci1.reset(), pc.StatusCode.SUCCESS)

    def test_commands(self):
        hci1.reset()

        self.assertEqual(hci1.set_tx_test_err_pattern(A32), pc.StatusCode.SUCCESS)
        self.assertIsNotNone(hci1.set_connection_op_flags(1, MAX_U32, True))

        key = list(secrets.token_bytes(32))
        self.assertEqual(hci1.set_256_priv_key(key), pc.StatusCode.SUCCESS)
        self.assertEqual(
            hci1.get_channel_map_periodic_scan_adv(1, False)[1],
            pc.StatusCode.SUCCESS,
        )
        self.assertIsNotNone(hci1.get_acl_test_report())

        self.assertEqual(
            hci1.set_local_num_min_used_channels(PhyOption.PHY_1M, 0, 10),
            pc.StatusCode.SUCCESS,
        )

        self.assertEqual(
            hci1.get_peer_min_num_channels_used(1)[1],
            pc.StatusCode.ERROR_CODE_UNKNOWN_CONN_ID,
        )

        self.assertEqual(
            hci1.set_validate_pub_key_mode(PubKeyValidateMode.ALT1),
            pc.StatusCode.SUCCESS,
        )

        addr, status = hci1.get_rand_address()

        self.assertTrue(addr > 0 or status != pc.StatusCode.SUCCESS)

        self.assertEqual(hci1.set_local_feature(0), pc.StatusCode.SUCCESS)

        self.assertEqual(hci1.set_operational_flags(0, True), pc.StatusCode.SUCCESS)

        self.assertEqual(
            hci1.set_encryption_mode(1, True, True),
            pc.StatusCode.ERROR_CODE_UNKNOWN_CONN_ID,
        )

        self.assertEqual(hci1.set_diagnostic_mode(True), pc.StatusCode.SUCCESS)

    def test_stats(self):
        stats, status = hci1.get_pdu_filter_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)

        stats, status = hci1.get_memory_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)

        stats, status = hci1.get_adv_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)
        stats, status = hci1.get_scan_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)

        stats, status = hci1.get_conn_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)

        stats, status = hci1.get_test_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)

        stats, status = hci1.get_aux_adv_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)

        stats, status = hci1.get_aux_scan_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)
        self.assertEqual(
            hci1.set_connection_phy_tx_power(1, 0, PhyOption.PHY_1M),
            pc.StatusCode.ERROR_CODE_UNKNOWN_CONN_ID,
        )


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""
hello_hci.py

Description: Bare minimum example of using the BleHci class

"""
from max_ble_hci import BleHci

# Path to serial port used for HCI
PORT = ""

controller = BleHci(PORT)
event = controller.reset()

print(event)



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""
Module contains definitions for ADI vendor-specific HCI commands.
"""
# pylint: disable=too-many-lines, too-many-arguments, too-many-public-methods
from typing import Dict, List, Optional, Tuple, Union

from ._hci_logger import get_formatted_logger
from ._transport import SerialUartTransport
from .constants import MAX_U32, MAX_U64, PayloadOption, PhyOption, PubKeyValidateMode
from .data_params import (
    AdvPktStats,
    DataPktStats,
    MemPktStats,
    PduPktStats,
    PoolStats,
    ScanPktStats,
    TestReport,
)
from .hci_packets import CommandPacket, EventPacket, byte_length
from .packet_codes import StatusCode
from .packet_defs import OCF, OGF
from .utils import to_le_nbyte_list


class VendorSpecificCmds:
    """Definitions for ADI vendor-specific HCI commands.

    Class contains functions used to implement Analog Devices
    vendor-specific HCI commands. Used as a parent for the full
    Analog Devices BLE HCI class.

    Arguments
    ---------
    port : SerialUartTransport
        Serial port interfacing object.
    logger_name: str
        Name used to reference the HCI logger.

    Attributes
    ----------
    port : SerialUartTransport
        Serial port interfacing object.
    logger : logging.Logger
        HCI logging object referenced by the `name` argument.

    """

    def __init__(self, port: SerialUartTransport, logger_name: str):
        self.port = port
        self.logger = get_formatted_logger(name=logger_name)

    def send_vs_command(
        self, ocf: OCF, params: List[int] = None, return_evt: bool = False
    ) -> Union[EventPacket, StatusCode]:
        """Send a vendor-specific command to the test board.

        Sends a command from the OGF Vendor Specific subgroup to the DUT.

        Parameters
        ----------
        ocf : OCF
            Opcode command field value for the desired HCI command.
        params : List[int], optional
            Command parameters as single-byte values.
        return_evt : bool, optional
            If true, function returns full `EventPacket` object. If
            false, function returns only the status code.

        Returns
        -------
        Union[StatusCode, EventPacket]
            If `return_evt` argument is true, the full return packet
            from the DUT. If `return_evt` argument is false, the return
            packet status code.


        """
        cmd = CommandPacket(OGF.VENDOR_SPEC, ocf, params=params)
        if return_evt:
            return self.port.send_command(cmd)

        return self.port.send_command(cmd).status

    def set_address(self, addr: int) -> StatusCode:
        """Sets the BD address.

        Function sets the chip BD address. Address can be given
        as either a bytearray or as a list of integer values.

        Parameters
        ----------
        addr : List[int]
            Desired BD address.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        params = to_le_nbyte_list(addr, 6)
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_BD_ADDR, params=params)

    def reset_connection_stats(self) -> StatusCode:
        """Reset accumulated connection stats.

        Sends a vendor-specific command to the DUT, telling it to
        reset all accumulated connection statisitics.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        return self.send_vs_command(OCF.VENDOR_SPEC.RESET_CONN_STATS)

    def enable_autogenerate_acl(self, enable: bool) -> StatusCode:
        """Enable/disable automatic generation of ACL packets.

        Sends a vendor-specific command to the DUT, telling it to
        enable or disable automatic generation of asynchronous
        connection-less packets.

        Parameters
        ----------
        enable: bool
            Enable automatic ACL packet generation?

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        return self.send_vs_command(
            OCF.VENDOR_SPEC.ENA_AUTO_GEN_ACL, params=int(enable)
        )

    def generate_acl(
        self, handle: int, packet_len: int, num_packets: int
    ) -> StatusCode:
        """Command board to generate ACL data.

        Sends a vendor-specific command to the DUT telling it
        to generate/send ACL data in accordance with the provided
        packet length and number of packets. A test end function
        must be called to end this process on the board.

        Parameters
        ----------
        handle : int
            Connection handle.
        packet_len : int
            Desired packet length.
        num_packets : int
            Desired number of packets to send.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `handle` is larger than 2 bytes.
        ValueError
            If `packet_len` is greater than 65535.
        ValueError
            If `num_packets` is greater than 255.

        """
        if byte_length(handle) > 2:
            raise ValueError(
                f"Handle ({handle}) is too large, must be 2 bytes or less."
            )
        if num_packets > 0xFFFF:
            raise ValueError(
                f"Num packets too large ({num_packets}), must be 65535 or less."
            )
        if packet_len > 0xFF:
            raise ValueError(
                f"Packet length too large ({packet_len}), must be 255 or less."
            )

        params = to_le_nbyte_list(handle, 2)
        params.append(packet_len)
        params.extend(to_le_nbyte_list(num_packets, 2))
        return self.send_vs_command(OCF.VENDOR_SPEC.GENERATE_ACL, params=params)

    def enable_acl_sink(self, enable: bool) -> StatusCode:
        """Enable/disable ACL sink.

        Sends a vendor-specific command to the DUT, telling it to
        enable or disable asynchronous connection-less packet sink.

        Parameters
        ----------
        enable : bool
            Enable ACL sink?

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        params = int(enable)
        return self.send_vs_command(OCF.VENDOR_SPEC.ENA_ACL_SINK, params=params)

    def tx_test_vs(
        self,
        channel: int = 0,
        phy: Union[PhyOption, int] = PhyOption.PHY_1M,
        payload: Union[PayloadOption, int] = PayloadOption.PLD_PRBS15,
        packet_len: int = 0,
        num_packets: int = 0,
    ) -> StatusCode:
        """Start a vendor-specific transmitter test.

        Sends a vendor-specific command to the DUT, telling it to
        start a DTM transmitter test in accordance with the given
        parameters.

        Parameters
        ----------
        channel : int
            The channel on which transmission should take place.
        phy : Union[PhyOption, int]
            The PHY that should be used by the transmitter.
        payload : Union[PayloadOption, int]
            The packet payload type that should be transmitted.
        packet_len : int
            The desired length of the transmitted packets.
        num_packets : int
            The number of packets to transmit. Set to `0` to
            enable continuous transmission.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `channel` is greater than 39 or less than 0.
        ValueError
            If `packet_len` is greater than 255.
        ValueError
            If `num_packets` is greater than 65535.

        """
        if not 0 <= channel < 40:
            raise ValueError(
                f"Channel out of bandwidth ({channel}), must be in range [0, 40)."
            )
        if packet_len > 0xFF:
            raise ValueError(
                f"Packet length too large ({packet_len}), must be 255 or less."
            )
        if num_packets > 0xFFFF:
            raise ValueError(
                f"Num packets too large ({num_packets}), must be 65535 or less."
            )

        if isinstance(payload, PayloadOption):
            payload = payload.value
        if isinstance(phy, PhyOption):
            phy = phy.value

        params = [channel, packet_len, payload, phy]
        params.extend(to_le_nbyte_list(num_packets, 2))
        return self.send_vs_command(OCF.VENDOR_SPEC.TX_TEST, params=params)

    def rx_test_vs(
        self,
        channel: int = 0,
        phy: Union[PhyOption, int] = PhyOption.PHY_1M,
        num_packets: int = 0,
        modulation_idx: int = 0,
    ) -> StatusCode:
        """Start a vendor-specific receiver test.

        Sends a vendor-specific command to the DUT, telling it to
        start a DTM receiver test in accordance with the given
        parameters.

        Parameters
        ----------
        channel : int
            The channel on which the receiver should listen for packets.
        phy : Union[PhyOption, int]
            The PHY that should be used by the receiver.
        num_packets : int
            The number of packets that the receiver is expected to receive,
            i.e. the number of packets the transmitter is sending.
        modulation_idx : int
            The expected modulation index of the transmitter. Indicates
            whether the modulation index is standard (0) or stable (1).

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `channel` is greater than 39 or less than 0.
        ValueError
            if `num_packets` is greater than 65535.

        """
        if not 0 <= channel < 40:
            raise ValueError(
                f"Channel out of bandwidth ({channel}), must be in range [0, 40)."
            )
        if num_packets > 0xFFFF:
            raise ValueError(
                f"Num packets too large ({num_packets}), must be 65535 or less."
            )

        if isinstance(phy, PhyOption):
            phy = phy.value

        params = [channel, phy, modulation_idx]
        params.extend(to_le_nbyte_list(num_packets, 2))
        return self.send_vs_command(OCF.VENDOR_SPEC.RX_TEST, params=params)

    def reset_test_stats(self) -> StatusCode:
        """Reset accumulated test stats.

        Sends a vendor-specific command to the DUT, telling it to
        reset all accumulated test statistics.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        return self.send_vs_command(OCF.VENDOR_SPEC.RESET_TEST_STATS)

    def set_adv_tx_power(self, tx_power: int) -> StatusCode:
        """Set the advertising TX power.

        Sends a vendor-specific command to the DUT, telling it to
        set the advertising TX power in accordance with the given
        value.

        Parameters
        ----------
        tx_power : int
            Desired advertising TX power.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `tx_power` is greater than 127 or less than -127.

        """
        if not -127 < tx_power < 127:
            raise ValueError(
                f"TX power ({tx_power}) out of range, must be in range [-127, 127]."
            )

        return self.send_vs_command(OCF.VENDOR_SPEC.SET_ADV_TX_PWR, params=tx_power)

    def set_conn_tx_power(self, tx_power: int, handle: int = 0x0000) -> StatusCode:
        """Set the connection TX power.

        Sends a vendor-specific command to the DUT, telling it to
        set the TX power on the indicated connection in accordance
        with the given value.

        Parameters
        ----------
        tx_power : int
            Desired connection TX power.
        handle : int, optional
            The handle to the desired connection.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `handle` is more than 2 bytes in size.
        ValueError
            If `tx_power` is greater than 127 or less than -127.

        """
        if byte_length(handle) > 2:
            raise ValueError(
                f"Handle ({handle}) is too large, must be 2 bytes or less."
            )
        if not -127 < tx_power < 127:
            raise ValueError(
                f"TX power ({tx_power}) out of range, must be in range [-127, 127]."
            )

        params = to_le_nbyte_list(handle, 2)
        params.append(tx_power)
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_CONN_TX_PWR, params=params)

    def set_channel_map(
        self,
        channels: Optional[Union[List[int], int]] = None,
        handle: int = 0x0000,
    ) -> StatusCode:
        """Set the channel map for an existing connection.

        Sends a vendor-specific command to the DUT, telling it to
        set the channel map for the indicated connection in
        accordance with the mask generated from the given channel
        values.

        Parameters
        ----------
        channels : Union[List[int], int], optional
            The channel(s) that should be included in the connection
            channel map.
        handle : int, optional
            The handle to the desired connection.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `handle` is more than 2 bytes in size.

        """
        if byte_length(handle) > 2:
            raise ValueError(
                f"Handle ({handle}) is too large, must be 2 bytes or less."
            )

        if channels:
            channels = channels if isinstance(channels, list) else [channels]
            channel_mask = 0x0000000000
            for chan in channels:
                if chan in [37, 38, 39]:
                    continue
                channel_mask = channel_mask | (1 << chan)
        else:
            channel_mask = 0x1FFFFFFFFF

        params = to_le_nbyte_list(handle, 2)
        params.extend(to_le_nbyte_list(channel_mask, 5))

        return self.send_vs_command(OCF.VENDOR_SPEC.SET_CHAN_MAP, params=params)

    def read_register(
        self, addr: int, length: int, print_data: bool = False
    ) -> Tuple[List[int], StatusCode]:
        """Read a number of bytes from a register.

        Sends a vendor-specific command to the DUT, telling it to
        read bytes from a register in accordance with the given
        length and register address values.

        Parameters
        ----------
        addr : int
            The address at which the read should begin.
        length : int
            The number of bytes to read.
        print_data : bool, optional
            Print read data to the console?

        Returns
        -------
        List[int]
            The read data.
        StatusCode
            The return packet status code.

        """
        params = [length]
        params.extend(to_le_nbyte_list(addr, 4))
        evt = self.send_vs_command(
            OCF.VENDOR_SPEC.REG_READ, params=params, return_evt=True
        )

        param_lens = [4] * (length // 4)
        if not length % 4 == 0:
            param_lens.append(length % 4)
        read_data = evt.get_return_params(param_lens=param_lens)

        if print_data:
            for idx, plen in enumerate(param_lens):
                if plen == 1:
                    self.logger.info("0x%08X: 0x______%02X", addr, read_data[idx])
                elif plen == 2:
                    self.logger.info("0x%08X: 0x____%04X", addr, read_data[idx])
                elif plen == 3:
                    self.logger.info("0x%08X: 0x__%06X", addr, read_data[idx])
                else:
                    self.logger.info("0x%08X: 0x%08X", addr, read_data[idx])

                addr += 4

        return read_data, evt.status

    def set_scan_channel_map(self, channel_map: int) -> StatusCode:
        """Set the channel map used for scanning.

        Sends a vendor-specific command to the DUT, telling it to
        set the channel map used for scanning in accordance with the
        given value.

        Parameters
        ----------
        channel_map : int
            Desired channel map to use for scanning.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_SCAN_CH_MAP, params=channel_map)

    def set_event_mask_vs(self, mask: int, enable: bool) -> StatusCode:
        """Enable/disable vendor specific events the board can generate.

        Sends a vendor-specific command to the DUT, telling it to
        enable/disable vendor-specific events that can be generated
        and returned to the host in accordance with the given mask.

        Parameters
        ----------
        mask : int
            Mask indicating the vendor-specific events that should
            be enabled/disabled. Events are indicated when their
            corresponding bit is set to `1`.
        enable : bool
            If true, enables the indicated events. If false, disables
            them.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        params = to_le_nbyte_list(mask, 8)
        params.append(int(enable))
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_EVENT_MASK, params=params)

    def set_tx_test_err_pattern(self, pattern: int) -> StatusCode:
        """Set the TX test mode error pattern.

        Sends a vendor-specific command to the DUT, telling it to
        set the pattern of errors for the TX test mode in accordance
        with the given value.

        Parameters
        ----------
        pattern : int
            Desired error pattern.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `pattern` is larger than 32 bits (4 bytes) in size.

        """

        if pattern > MAX_U32:
            raise ValueError(f"Pattern ({pattern}) too large, must be 32 bits or less.")

        params = to_le_nbyte_list(pattern, 4)
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_TX_TEST_ERR_PATT, params=params)

    def set_connection_op_flags(
        self, handle: int, flags: int, enable: bool
    ) -> StatusCode:
        """Set connection operational flags.

        Sends a vendor-specific command to the DUT, telling it to
        enable/disable the connection operational flags for the
        indicated connection in accordance with the values provided.

        Parameters
        ----------
        handle : int
            The handle to the desired connection.
        flags : int
            Mask indicating the desired connection operational flags
            that should be enabled/disabled. Flags are indicated when
            their corresponding bit is set to `1`.
        enable : bool
            If true, enables the indicated flags. If false, disables
            them.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `handle` is larger than 2 bytes in size.
        ValueError
            If `flags` is larger than 4 bytes in size.

        """
        if byte_length(handle) > 2:
            raise ValueError(
                f"Handle ({handle}) is too large, must be 2 bytes or less."
            )
        if byte_length(flags) > 4:
            raise ValueError(f"Flags ({flags}) is too large, must be 4 bytes or less.")

        params = to_le_nbyte_list(handle, 2)
        params.extend(to_le_nbyte_list(flags, 4))
        params.append(int(enable))
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_CONN_OP_FLAGS, params=params)

    def set_256_priv_key(self, priv_key: List[int]) -> StatusCode:
        """Set/clear the P-256 private key.

        Sends a vendor-specific command to the DUT, telling it to
        set or clear the P-256 private key used to generate key
        pairs and Diffie-hellman keys in accordance with the given
        value.

        Parameters
        ----------
        priv_key : list
            Desired P-256 private key. Setting to `0` will clear
            the key.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `priv_key` is larger than 32 bytes in size.

        """
        if len(priv_key) > 32:
            raise ValueError(
                f"Private key ({priv_key}) too large, must be 32 bytes or less."
            )

        return self.send_vs_command(
            OCF.VENDOR_SPEC.SET_P256_PRIV_KEY, params=priv_key[::-1]
        )

    def get_channel_map_periodic_scan_adv(
        self, handle: int, is_advertising: bool
    ) -> Tuple[int, StatusCode]:
        """Get the channel map used for periodic scanning/advertising.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve the channel map used for either periodic scanning
        or periodic advertising in accordance with the given values.

        Parameters
        ----------
        handle : int
            The handle to the desired periodic scanner/advertiser.
        is_advertising : bool
            Does the handle point to a periodic advertiser?

        Returns
        -------
        int
            The channel map returned by the DUT.
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `handle` is larger than 2 bytes in size.

        """
        if byte_length(handle) > 2:
            raise ValueError(
                f"Handle ({handle}) is too large, must be 2 bytes or less."
            )

        params = to_le_nbyte_list(handle, 2)
        params.append(int(is_advertising))
        evt = self.send_vs_command(
            OCF.VENDOR_SPEC.GET_PER_CHAN_MAP, params=params, return_evt=True
        )

        return evt.get_return_params(), evt.status

    def get_acl_test_report(self) -> Tuple[TestReport, StatusCode]:
        """Get ACL test report.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve the current ACL test report.

        Returns
        -------
        TestReport
            The ACL test report returned by the DUT.
        StatusCode
            The return packet status code.

        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_ACL_TEST_REPORT, return_evt=True)
        data = evt.get_return_params(param_lens=[4, 4, 4, 4])

        stats = TestReport(
            rx_pkt_count=data[0],
            rx_oct_count=data[1],
            gen_pkt_count=data[2],
            gen_oct_count=data[3],
        )

        return stats, evt.status

    def set_local_num_min_used_channels(
        self, phy: PhyOption, pwr_thresh: int, min_used: int
    ) -> StatusCode:
        """Set local minimum number of used channels.

        Sends a vendor-specific command to the DUT, telling it to
        set the local minimum number of used channels in accordance
        with the given PHY, power threshold, and minimum values.

        Parameters
        ----------
        phy : PhyOption
            PHY on which the process should take place.
        pwr_thresh : int
            Power threshold for the selected PHY.
        min_used : int
            Minimum number of used channels.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `pwr_thresh` is greater than 127 or less than -127.
        ValueError
            if `min_used` is greater than 37 or less than 1.

        """
        if not -127 < pwr_thresh < 127:
            raise ValueError(
                f"Thresh ({pwr_thresh}) out of range, must be in range [-127, 127]."
            )
        if not 0 < min_used <= 37:
            raise ValueError(
                f"Min used ({min_used}) out of range, must be in range [1, 37]."
            )

        if phy == PhyOption.PHY_CODED_S2:
            phy = PhyOption.PHY_CODED

        params = [phy.value, pwr_thresh, min_used]
        return self.send_vs_command(
            OCF.VENDOR_SPEC.SET_LOCAL_MIN_USED_CHAN, params=params
        )

    def get_peer_min_num_channels_used(
        self, handle: int
    ) -> Tuple[Dict[PhyOption, int], StatusCode]:
        """Get the minimum number of channels used by a peer.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve the minimum number of channels used by a peer
        device as indicated by the given value.

        Parameters
        ----------
        handle : int
            Handle to the desired peer connection.

        Returns
        -------
        Dict[PhyOption, int]
            Peer minimum number of used channels by PHY type.
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `handle` is larger than 2 bytes in size.

        """
        if byte_length(handle) > 2:
            raise ValueError(
                f"Handle ({handle}) is too large, must be 2 bytes or less."
            )

        params = to_le_nbyte_list(handle, 2)
        evt = self.send_vs_command(
            OCF.VENDOR_SPEC.GET_PEER_MIN_USED_CHAN, params=params, return_evt=True
        )
        data = evt.get_return_params(param_lens=[1, 1, 1])

        min_used_map = {
            PhyOption.PHY_1M: data[0],
            PhyOption.PHY_2M: data[1],
            PhyOption.PHY_CODED: data[2],
        }

        return min_used_map, evt.status

    def set_validate_pub_key_mode(self, mode: PubKeyValidateMode) -> StatusCode:
        """Set the mode used to validate the public key.

        Sends a vendor-specific command to the DUT, telling it to
        set the mode used to validate the public key in accordance
        with the given value.

        Parameters
        ----------
        mode : PubKeyValidateMode
            Desired public key validation mode.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        return self.send_vs_command(
            OCF.VENDOR_SPEC.VALIDATE_PUB_KEY_MODE, params=[mode.value]
        )

    def get_rand_address(self) -> Tuple[int, StatusCode]:
        """Get a random device address.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve a random device address.

        Returns
        -------
        int
            Random device address retrieved by the DUT.
        StatusCode
            The return packet status code.

        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_RAND_ADDR, return_evt=True)

        return evt.get_return_params(), evt.status

    def set_local_feature(self, features: int) -> StatusCode:
        """Set local supported features.

        Sends a vendor-specific command to the DUT, telling it to
        set the local supported features in accordance with the
        given value.

        Parameters
        ----------
        features : int
            Mask indicating the local supported features. Setting
            a bit to `1` will enable the indicated feature. Setting
            a bit to `0` will disable it.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `features` is larger than 64 bits (8 bytes) in size.

        """
        if features > MAX_U64:
            raise ValueError(
                f"Feature mask ({features}) is too large, must be 64 bits or less."
            )

        params = to_le_nbyte_list(features, 8)
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_LOCAL_FEAT, params=params)

    def set_operational_flags(self, flags: int, enable: bool) -> StatusCode:
        """Enable/disable operational flags.

        Sends a vendor-specific command to the DUT, telling it to
        enable or disable operational flags in accordance with the
        values provided.

        Parameters
        ----------
        flags : int
            Mask indicating the desired operational flags that should
            be enabled/disabled. Flags are indicated when their
            corresponding bit is set to `1`.
        enable : bool
            If true, enables the indicated flags. If false, disabled
            them.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `flags` is larger than 32 bits (4 bytes) in size.

        """
        if flags > MAX_U32:
            raise ValueError(f"Flags ({flags}) is too large, must be 32 bits or less.")

        params = to_le_nbyte_list(flags, 4)
        params.append(int(enable))
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_OP_FLAGS, params=params)

    def get_pdu_filter_stats(self) -> Tuple[PduPktStats, StatusCode]:
        """Get the accumulated PDU filter stats.

        Sends a vendor-specific command to the DUT, telling it to
        retrieves the current accumulated PDU filter statistics.

        Returns
        -------
        PduPktStats
            PDU filter statistics report returned by the DUT.
        StatusCode
            The return packet status code.

        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_PDU_FILT_STATS, return_evt=True)
        data = evt.get_return_params(param_lens=[2] * 19)

        stats = PduPktStats(
            fail_pdu=data[0],
            pass_pdu=data[1],
            fail_whitelist=data[2],
            pass_whitelist=data[3],
            fail_peer_addr_match=data[4],
            pass_peer_addr_match=data[5],
            fail_local_addr_match=data[6],
            pass_local_addr_match=data[7],
            fail_peer_rpa_verify=data[8],
            pass_peer_rpa_verify=data[9],
            fail_local_rpa_verify=data[10],
            pass_local_rpa_verify=data[11],
            fail_peer_priv_addr=data[12],
            fail_local_priv_addr=data[13],
            fail_peer_addr_res_req=data[14],
            pass_peer_addr_res_req=data[15],
            pass_local_addr_res_opt=data[16],
            peer_res_addr_pend=data[17],
            local_res_addr_pend=data[18],
        )

        return stats, evt.status

    def set_encryption_mode(
        self, handle: int, enable: bool, nonce_mode: bool
    ) -> StatusCode:
        """Set the encryption mode of an existing connection.

        Sends a vendor-specific command to the DUT, telling it to
        set the encryption mode of the indicated connection in
        accordance with the values provided.

        Parameters
        ----------
        handle : int
            Handle to the desired connection.
        enable : bool
            Enable authentication?
        nonce_mode : bool
            Enable nonce mode?

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `handle` is larger than 2 bytes in size.

        """
        if byte_length(handle) > 2:
            raise ValueError(
                f"Handle ({handle}) is too large, must be 2 bytes or less."
            )

        params = [int(enable)]
        params.append(int(nonce_mode))
        params.extend(to_le_nbyte_list(handle, 2))
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_ENC_MODE, params=params)

    def set_diagnostic_mode(self, enable: bool) -> StatusCode:
        """Enable/disable diagnostic mode.

        Sends a vendor-specific command to the DUT, telling it to
        enable or disable the PAL system assert trap in accordance
        with the provided value.

        Parameters
        ----------
        enable : bool
            Enable diagnostic mode?

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_DIAG_MODE, params=int(enable))

    def enable_sniffer_packet_forwarding(self, enable: bool) -> StatusCode:
        """Enable/disable sniffer packet forwarding.

        Sends a vendor-specific command to the DUT, telling it to
        enable or disable sniffer packet forwarding in accordance
        with the value provided.

        Parameters
        ----------
        enable : bool
            Enable sniffer packet forwarding?

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        out_method = 0  # HCI through tokens, only available option
        params = [out_method, int(enable)]
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_SNIFFER_ENABLE, params=params)

    def get_memory_stats(self) -> Tuple[MemPktStats, StatusCode]:
        """Get memory and system stats.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve the current memory and system statistics.

        Returns
        -------
        MemPktStats
            Memory and system statistics report retrieved by the DUT.
        StatusCode
            The return packet status code.

        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_SYS_STATS, return_evt=True)
        data = evt.get_return_params(
            param_lens=[2, 2, 4, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
        )

        stats = MemPktStats(
            stack=data[0],
            sys_assert_cnt=data[1],
            free_mem=data[2],
            used_mem=data[3],
            max_connections=data[4],
            conn_ctx_size=data[5],
            cs_watermark_lvl=data[6],
            ll_watermark_lvl=data[7],
            sch_watermark_lvl=data[8],
            lhci_watermark_lvl=data[9],
            max_adv_sets=data[10],
            adv_set_ctx_size=data[11],
            ext_scan_max=data[12],
            ext_scan_ctx_size=data[13],
            max_num_ext_init=data[14],
            ext_init_ctx_size=data[15],
            max_per_scanners=data[16],
            per_scan_ctx_size=data[17],
            max_cig=data[18],
            cig_ctx_size=data[19],
            cis_ctx_size=data[20],
        )

        return stats, evt.status

    def get_adv_stats(self) -> Tuple[AdvPktStats, StatusCode]:
        """Get the accumulated advertising stats.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve the current accumulated advertising statistics.

        Returns
        -------
        AdvPktStats
            Advertising statistics report retrieved by the DUT.
        StatusCode
            The return packet status code.

        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_ADV_STATS, return_evt=True)
        data = evt.get_return_params(param_lens=[4, 4, 4, 2, 4, 4, 2, 2, 2, 2])

        stats = AdvPktStats(
            tx_adv=data[0],
            rx_req=data[1],
            rx_req_crc=data[2],
            rx_req_timeout=data[3],
            tx_resp=data[4],
            err_adv=data[5],
            rx_setup=data[6],
            tx_setup=data[7],
            rx_isr=data[8],
            tx_isr=data[9],
        )

        return stats, evt.status

    def get_scan_stats(self) -> Tuple[ScanPktStats, StatusCode]:
        """Get Scan stats

        Returns
        -------
        Tuple[ScanPktStats, StatusCode]
            Accumulated scanning stats and status code
        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_SCAN_STATS, return_evt=True)
        data = evt.get_return_params(param_lens=[4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2])

        stats = ScanPktStats(
            rx_adv=data[0],
            rx_adv_crc=data[1],
            rx_adv_timeout=data[2],
            tx_req=data[3],
            rx_rsp=data[4],
            rx_rsp_crc=data[5],
            rx_rsp_timeout=data[6],
            err_scan=data[7],
            rx_setup=data[8],
            tx_setup=data[9],
            rx_isr=data[10],
            tx_isr=data[11],
        )

        return stats, evt.status

    def get_conn_stats(self) -> Tuple[DataPktStats, StatusCode]:
        """Get the stats captured during a connection.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve the statistics captured during a connection.

        Returns
        -------
        DataPktStats
            Connection statistics report retrieved by the DUT.
        StatusCode
            The return packet status code.

        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_CONN_STATS, return_evt=True)
        data = evt.get_return_params(param_lens=[4, 4, 4, 4, 4, 2, 2, 2, 2])

        stats = DataPktStats(
            rx_data=data[0],
            rx_data_crc=data[1],
            rx_data_timeout=data[2],
            tx_data=data[3],
            err_data=data[4],
            rx_setup=data[5],
            tx_setup=data[6],
            rx_isr=data[7],
            tx_isr=data[8],
        )

        return stats, evt.status

    def get_test_stats(self) -> Tuple[DataPktStats, StatusCode]:
        """Get the stats captured during test mode.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve the statistics captured during DTM.

        Returns
        -------
        DataPktStats
            Test mode statistics report retrieved by the DUT.
        StatusCode
            The return packet status code.

        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_TEST_STATS, return_evt=True)
        data = evt.get_return_params(param_lens=[4, 4, 4, 4, 4, 2, 2, 2, 2])

        stats = DataPktStats(
            rx_data=data[0],
            rx_data_crc=data[1],
            rx_data_timeout=data[2],
            tx_data=data[3],
            err_data=data[4],
            rx_setup=data[5],
            tx_setup=data[6],
            rx_isr=data[7],
            tx_isr=data[8],
        )

        return stats, evt.status

    def get_pool_stats(self) -> Tuple[List[PoolStats], StatusCode]:
        """Get the memory pool stats captured during runtime.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve the memory pool statistics captured during runtime.

        Returns
        -------
        List[PoolStats]
            Memory pool statistics reports retrieved by the DUT.
        StatusCode
            The return packet status code.

        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_POOL_STATS, return_evt=True)
        num_pools = evt.evt_params[0]

        param_lens = [1]
        param_lens.extend([2, 1, 1, 1, 2] * num_pools)

        data = evt.get_return_params(param_lens=param_lens, use_raw=True)

        stats = []
        num_pools = data.pop(0)
        for _ in range(num_pools):
            stats.append(
                PoolStats(
                    buf_size=data.pop(0),
                    num_buf=data.pop(0),
                    num_alloc=data.pop(0),
                    max_alloc=data.pop(0),
                    max_req_len=data.pop(0),
                )
            )

        return stats, evt.status

    def set_additional_aux_ptr_offset(self, delay: int, handle: int) -> StatusCode:
        """Set auxiliary packet offset delay.

        Sends a vendor-specific command to the DUT, telling it to
        set the auxiliary packet offset delay in accordance with
        the given values.

        Parameters
        ----------
        delay : int
            Desired delay. Set to 0 to disable.
        handle : int
            Handle to the desired connection.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `delay` is larger than 4 bytes in size.

        """
        if byte_length(delay) > 4:
            raise ValueError(f"Delay ({delay}) is too large, must be 4 bytes or less.")

        params = to_le_nbyte_list(delay, 4)
        params.append(handle)
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_AUX_DELAY, params=params)

    def set_ext_adv_data_fragmentation(
        self, handle: int, frag_length: int
    ) -> StatusCode:
        """Set the extended advertising fragmentation length.

        Sends a vendor-specific command to the DUT, telling it to
        set the extended advertising fragmentation length in
        accordance with the values provided.

        Parameters
        ----------
        handle : int
            Desired advertising handle.
        frag_length : int
            Desired fragmentation length.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        params = [handle, frag_length]
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_EXT_ADV_FRAG_LEN, params=params)

    def set_extended_advertising_phy_opts(
        self, handle: int, primary: int, secondary: int
    ) -> StatusCode:
        """Set extended advertising PHY options.

        Sends a vendor-specific command to the DUT, telling it to
        set the extended advertising PHY options in accordance with
        the values provided.

        Parameters
        ----------
        handle : int
            Desired advertising handle.
        primary : int
            Desired primary advertising channel PHY options.
        secondary : int
            Desired secondary advertising channel PHY options.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        params = [handle, primary, secondary]
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_EXT_ADV_PHY_OPTS, params=params)

    def set_extended_advertising_default_phy_opts(self, phy_opts: int) -> StatusCode:
        """Set the extended advertising default TX PHY options.

        Sends a vendor-specific command to the DUT, telling it to
        set the default TX PHY options for the extended advertising
        slave primary and secondary channels in accordance with the
        value provided.

        Parameters
        ----------
        phy_opts : int
            Desired PHY options.

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        return self.send_vs_command(
            OCF.VENDOR_SPEC.SET_EXT_ADV_DEF_PHY_OPTS, params=phy_opts
        )

    def generate_iso_packets(
        self, handle: int, packet_len: int, num_packets: int
    ) -> StatusCode:
        """Generate ISO packets.

        Sends a vendor-specific command to the DUT, telling it to
        generate ISO packets on the indicated connection in accordance
        with the parameters provided.

        Parameters
        ----------
        handle : int
            Handle to the desired connection.
        packet_len : int
            Desired packet length.
        num_packets : int
            Number of ISO packets to send.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `handle` is larger than 2 bytes in size.
        ValueError
            If `packet_len` is larger than 2 bytes in size.

        """
        if byte_length(handle) > 2:
            raise ValueError(
                f"Handle ({handle}) is too large, must be 2 bytes or less."
            )
        if byte_length(packet_len) > 2:
            raise ValueError(
                f"Packet length ({packet_len}) is too large, must be 2 bytes or less."
            )

        params = to_le_nbyte_list(handle, 2)
        params.extend(to_le_nbyte_list(packet_len, 2))
        params.append(num_packets)
        return self.send_vs_command(OCF.VENDOR_SPEC.GENERATE_ISO, params=params)

    def get_iso_test_report(self) -> Tuple[TestReport, StatusCode]:
        """Get the stats collected during an ISO test.

        Sends a vendor-specific command to the DUT, telling it to
        retrieves the statistics collected during an ISO test.

        Returns
        -------
        TestReport
            The ISO test statistics report retrieved by the DUT.
        StatusCode
            The return packet status code.

        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_ISO_TEST_REPORT, return_evt=True)
        data = evt.get_return_params(param_lens=[4, 4, 4, 4])

        stats = TestReport(
            rx_pkt_count=data[0],
            rx_oct_count=data[1],
            gen_pkt_count=data[2],
            gen_oct_count=data[3],
        )

        return stats, evt.status

    def enable_iso_packet_sink(self, enable: bool) -> StatusCode:
        """Enable/disable ISO packet sink.

        Sends a vendor-specific command to the DUT, telling it to
        enable or disable ISO packet sink in accordance with the
        value provided.

        Parameters
        ----------
        enable : bool
            Enable ISO packet sink?

        Returns
        -------
        StatusCode
            The return packet status code.

        """
        return self.send_vs_command(OCF.VENDOR_SPEC.ENA_ISO_SINK, params=int(enable))

    def enable_autogen_iso_packets(self, packet_len: int) -> StatusCode:
        """Enable/disable automatic generation of ISO packets.

        Sends a vendor-specific command to the DUT, telling it to
        enable or disable the automatic generation of ISO packets
        in accordance with the values provided.

        Parameters
        ----------
        packet_len : int
            Desired ISO packet length. Set to 0 to disable automatic
            generation.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `packet_len` is larger than 32 bits (4 bytes) in size.

        """
        if packet_len > MAX_U32:
            raise ValueError(
                f"Packet length ({packet_len}) is too large, must be 4 bytes or less."
            )

        params = to_le_nbyte_list(packet_len, 4)
        return self.send_vs_command(OCF.VENDOR_SPEC.ENA_AUTO_GEN_ISO, params=params)

    def get_iso_connection_stats(self) -> Tuple[DataPktStats, StatusCode]:
        """Get the stats captured during an ISO connection.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve the statistics captured during an ISO connection.

        Returns
        -------
        DataPktStats
            The ISO connection statistics report retrieved by the DUT.
        StatusCode
            The return packet status code.

        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_ISO_TEST_REPORT, return_evt=True)
        data = evt.get_return_params(param_lens=[4, 4, 4, 4, 4, 2, 2, 2, 2])

        stats = DataPktStats(
            rx_data=data[0],
            rx_data_crc=data[1],
            rx_data_timeout=data[2],
            tx_data=data[3],
            err_data=data[4],
            rx_setup=data[5],
            tx_setup=data[6],
            rx_isr=data[7],
            tx_isr=data[8],
        )

        return stats, evt.status

    def get_aux_adv_stats(self) -> Tuple[AdvPktStats, StatusCode]:
        """Get the accumulated auxiliary advertising stats.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve the current accumulated auxiliary advertising
        statistics.

        Returns
        -------
        AdvPktStats
            The auxiliary advertising statistics report retrieved
            by the DUT.
        StatusCode
            The return packet status code.

        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_AUX_ADV_STATS, return_evt=True)
        data = evt.get_return_params(param_lens=[4, 4, 4, 2, 4, 4, 4, 2, 2, 2, 2])

        stats = AdvPktStats(
            tx_adv=data[0],
            rx_req=data[1],
            rx_req_crc=data[2],
            rx_req_timeout=data[3],
            tx_resp=data[4],
            tx_chain=data[5],
            err_adv=data[6],
            rx_setup=data[7],
            tx_setup=data[8],
            rx_isr=data[9],
            tx_isr=data[10],
        )

        return stats, evt.status

    def get_aux_scan_stats(self) -> Tuple[ScanPktStats, StatusCode]:
        """Get the accumulated auxiliary scan stats.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve the current accumulated auxiliary scan statistics.

        Returns
        -------
        ScanPktStats
            The auxiliary scan statistics report retrieved by the DUT.
        StatusCode
            The return packet status code.

        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_AUX_SCAN_STATS, return_evt=True)
        data = evt.get_return_params(
            param_lens=[4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2]
        )

        stats = ScanPktStats(
            rx_adv=data[0],
            rx_adv_crc=data[1],
            rx_adv_timeout=data[2],
            tx_req=data[3],
            rx_rsp=data[4],
            rx_rsp_crc=data[5],
            rx_rsp_timeout=data[6],
            rx_chain=data[7],
            rx_chain_crc=data[8],
            rx_chain_timeout=data[9],
            err_scan=data[10],
            rx_setup=data[11],
            tx_setup=data[12],
            rx_isr=data[13],
            tx_isr=data[14],
        )
        return stats, evt.status

    def get_periodic_scanning_stats(self) -> Tuple[ScanPktStats, StatusCode]:
        """Get the accumulated periodic scanning stats.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve the current accumulated periodic scanning statistics.

        Returns
        -------
        ScanPktStats
            The periodic scanning statistics report retrieved by the DUT.
        StatusCode
            The return packet status code.

        """
        evt = self.send_vs_command(OCF.VENDOR_SPEC.GET_PER_SCAN_STATS, return_evt=True)
        data = evt.get_return_params(param_lens=[4, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2])

        stats = ScanPktStats(
            rx_adv=data[0],
            rx_adv_crc=data[1],
            rx_adv_timeout=data[2],
            rx_chain=data[3],
            rx_chain_crc=data[4],
            rx_chain_timeout=data[5],
            err_scan=data[6],
            rx_setup=data[7],
            tx_setup=data[8],
            rx_isr=data[9],
            tx_isr=data[10],
        )
        return stats, evt.status

    def set_connection_phy_tx_power(
        self, handle: int, power: int, phy: PhyOption
    ) -> StatusCode:
        """Set the connection TX power level for a specific PHY.

        Sends a vendor-specific command to the DUT, telling it to
        set the connection TX power level for the indicated connection
        and PHY in accordance with the value provided.

        Parameters
        ----------
        handle : int
            Handle to the desired connection.
        power : int
            Desired TX power.
        phy : PhyOption
            PHY on which the TX power should be set.

        Returns
        -------
        StatusCode
            The return packet status code.

        Raises
        ------
        ValueError
            If `handle` is larger than 2 bytes in size.

        """
        if byte_length(handle) > 2:
            raise ValueError(
                f"Handle ({handle}) is too large, must be 2 bytes or less."
            )

        if phy == PhyOption.PHY_CODED_S2:
            phy = PhyOption.PHY_CODED

        params = to_le_nbyte_list(handle, 2)
        params.append(power)
        params.append(phy.value)
        return self.send_vs_command(OCF.VENDOR_SPEC.SET_CONN_PHY_TX_PWR, params=params)

    def get_rssi_vs(self, channel: int = 0) -> Tuple[int, StatusCode]:
        """Get the RSSI values.

        Sends a vendor-specific command to the DUT, telling it to
        retrieve the RSSI value for the indicated channel.

        Parameters
        ----------
        channel : int, optional
            Channel for which value should be retrieved.

        Returns
        -------
        int
            RSSI value for the indicated channel.
        StatusCode
            The return packet status code.


        Raises
        ------
        ValueError
            If channel is greater than 39 or less than 0.
        """
        if not 0 <= channel < 40:
            raise ValueError(
                f"Channel out of bandwidth ({channel}), must be in range [0, 40)."
            )

        evt = self.send_vs_command(
            OCF.VENDOR_SPEC.GET_RSSI, params=channel, return_evt=True
        )
        rssi = evt.get_return_params(signed=True)

        if rssi == -128:
            self.logger.warning("RSSI= -128, possible timeout occured")

        return rssi, evt.status

    def bb_enable(self) -> StatusCode:
        """Enable the baseband radio
        NOTE: Must be done before using RSSI
        Returns
        -------
        StatusCode
        """
        return self.send_vs_command(OCF.VENDOR_SPEC.BB_EN)

    def bb_disable(self) -> StatusCode:
        """Disable the baseband radio

        Returns
        -------
        StatusCode
        """
        return self.send_vs_command(OCF.VENDOR_SPEC.BB_DIS)



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""Module for HCI logger handling.

Module defines a logger for test host-controller interfaces.

Attributes
----------
hci_logger : logging.Logger
    Host-controller interface logging object.

"""
import logging
from typing import Dict


class _CustomFormatter(logging.Formatter):
    """Log message formatting class.

    Attributes
    ----------
    cyan : str
        ANSI color code representing the color cyan.
    white : str
        ANSI color code representing the color white.
    green : str
        ANSI color code representing the color green.
    yellow : str
        ANSI color code representing the color yellow.
    red : str
        ANSI color code representing the color red.
    bold_red : str
        ANSI color code representing bold red.
    reset : str
        ANSI reset code.
    format_str : str
        Default logger string format.
    format_info : str
        Logger string format for info-level messages.
    format_debug : str
        Logger string format for debug-level messages.
    FORMATS : Dict[int, str]
        Dictionary containing logger messaging formats.

    """

    cyan: str = "\x1b[36;20m"
    white: str = "\x1b[37;20m"
    green: str = "\x1b[32;20m"
    yellow: str = "\x1b[33;20m"
    red: str = "\x1b[31;20m"
    bold_red: str = "\x1b[31;1m"
    reset: str = "\x1b[0m"
    format_str: str = (
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    )
    format_info: str = "%(levelname)s - %(message)s"
    format_debug: str = "%(levelname)s - %(message)s"

    FORMATS: Dict[int, str] = {
        logging.DEBUG: white + format_debug + reset,
        logging.INFO: green + format_info + reset,
        logging.WARNING: yellow + format_str + reset,
        logging.ERROR: red + format_str + reset,
        logging.CRITICAL: bold_red + format_str + reset,
    }

    def format(self, record: logging.LogRecord) -> str:
        """Creates and returns formatted log message.

        Paremeters
        ----------
        record : logging.LogRecord
            Object containing the log record to be formatted.

        Returns
        -------
        str
            The formatted log message.

        """
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_formatted_logger(log_level=logging.INFO, name="BLE-HCI") -> logging.Logger:
    """Retrieves logger with basic custom format.

    The custom formatted logger applies basic coloring
    for logging of different levels.


    Parameters
    ----------
    log_level : int
        Any defined logging level such as logging.INFO.

    """
    logger = logging.getLogger(name)

    if logger.hasHandlers():
        return logger

    logger.setLevel(log_level)

    custom_handler = logging.StreamHandler()
    custom_handler.setLevel(log_level)
    custom_handler.setFormatter(_CustomFormatter())

    if not logger.handlers:
        logger.addHandler(custom_handler)

    return logger



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""Contains full HCI implementation."""
# pylint: disable=too-many-arguments
import logging
from typing import Any, Callable, Optional, Union

from ._hci_logger import get_formatted_logger
from ._transport import SerialUartTransport
from .ble_standard_cmds import BleStandardCmds
from .constants import ADI_PORT_BAUD_RATE
from .data_params import AdvParams, ConnParams
from .hci_packets import AsyncPacket, CommandPacket, EventPacket
from .packet_codes import StatusCode
from .vendor_spec_cmds import VendorSpecificCmds


class BleHci(BleStandardCmds, VendorSpecificCmds):
    """Host-controller interface.

    The BleHci object defines a host-controller interface for
    BLE testing on any BLE-compatible microchip. Controller provides
    implementations for both BLE standard HCI command and ADI vendor
    specific commands. Support is also provided for the creation and
    use of custom vendor-specific commands.

    Parameters
    ----------
    port_id : str
        ID string for the port on which a connection should be
        established.
    baud : int
        Port baud rate.
    id_tag : str
        Connection ID string to use when logging.
    log_level : Union[str, int]
        Logging level.
    logger_name : str
        Name that should be used to reference HCI logger.
    retries : int
        Number of times a port read should be retried before
        and error is thrown.
    timeout : float
        Port timeout.
    async_callback : Callable[[AsyncPacket], Any], optional
        Function pointer defining the process that should be taken
        when an async packet is received. If not defined, the async
        packet will be thrown out.
    evt_callback : Callable[[EventPacket], Any], optional
        Function pointer defining the process that should be taken
        when an unexpected event packet is received. If not defined,
        the event packet will be thrown out.

    Attributes
    ----------
    port_id : str
        Id string for the port on which a connection has been
        established
    port : SerialUartTransport
        Serial port interfacing object connected to the DUT.
    id_tag : str
        Connection ID string used by the logger.
    logger : logging.Logger
        HCI logging object reference by the `logger_name` argument.
    retries : int
        Number of times a port read should be retried before an error
        is thrown.
    timeout : float
        Port timeout.

    """

    def __init__(
        self,
        port_id: str,
        baud: int = ADI_PORT_BAUD_RATE,
        id_tag: str = "DUT",
        log_level: Union[str, int] = "INFO",
        logger_name: str = "BLE-HCI",
        retries: int = 0,
        timeout: float = 1.0,
        async_callback: Optional[Callable[[AsyncPacket], Any]] = None,
        evt_callback: Optional[Callable[[EventPacket], Any]] = None,
    ):
        self.port_id = port_id
        self.port = None
        self.id_tag = id_tag
        self.logger = get_formatted_logger(log_level=log_level, name=logger_name)
        self.retries = retries
        self.timeout = timeout
        self._init_ports(port_id, baud, logger_name, async_callback, evt_callback)
        super().__init__(self.port, logger_name)

    def get_log_level(self) -> str:
        """Retrieve the current log level.

        Retrieved the current logging level in string
        format.

        Returns
        -------
        str
            The current logging level.

        """
        level = self.logger.level
        if level == logging.DEBUG:
            return "DEBUG"
        if level == logging.INFO:
            return "INFO"
        if level == logging.WARNING:
            return "WARNING"
        if level == logging.ERROR:
            return "ERROR"
        if level == logging.CRITICAL:
            return "CRITICAL"
        return "NOTSET"

    def set_log_level(self, level: Union[str, int]) -> None:
        """Sets log level.

        Provides intermediary control over the logging level
        of the host-controller interface module logger. If
        necessary, desired log level is automatically converted
        from a string to an integer. As such, both strings and
        integers are valid inputs to the `level` parameter.

        Parameters
        ----------
        level : Union[int, str]
            Desired log level.

        """
        if isinstance(level, int):
            self.logger.setLevel(level)
            return

        ll_str = level.upper()
        if ll_str == "DEBUG":
            self.logger.setLevel(logging.DEBUG)
        elif ll_str == "INFO":
            self.logger.setLevel(logging.INFO)
        elif ll_str == "WARNING":
            self.logger.setLevel(logging.WARNING)
        elif ll_str == "ERROR":
            self.logger.setLevel(logging.ERROR)
        elif ll_str == "CRITICAL":
            self.logger.setLevel(logging.CRITICAL)
        else:
            self.logger.setLevel(logging.NOTSET)
            self.logger.warning(
                "Invalid log level string: %s, level set to 'logging.NOTSET'", ll_str
            )

    def start_advertising(
        self, connect: bool = True, adv_params: Optional[AdvParams] = None
    ) -> StatusCode:
        """Start advertising.

        Convenience function which sends a sequence of commands to
        the DUT, telling it to begin advertising. PHYs preferences
        cannot be set when using this function, but advertising
        parameters can be using the optional `adv_params` parameter.
        If a value is provided for `adv_params`, the value of the
        `connect` parameter is ignored. If no value is provided, all
        advertising parameters are defaulted and `connect` is used
        to determine the advertising type.

        Parameters
        ----------
        connect : bool, optional
            Make connectable? If true, advertising type is set to
            `0x0 (ADV_IND)`. If false, advertising type is set to
            `0x3 (ADV_NONCONN_IND)`. Ignored if `adv_params` is
            not None.
        adv_params : AdvParams, optional.
            Advertising parameters.

        Returns
        -------
        StatusCode
            The return status of the enable advertising command.

        """

        self.reset_connection_stats()
        self.set_default_phy(all_phys=0, tx_phys=7, rx_phys=7)

        if adv_params is None:
            adv_type = 0 if connect else 3
            adv_params = AdvParams(adv_type=adv_type)
        self.set_adv_params(adv_params)

        status = self.enable_adv(True)

        return status

    def init_connection(
        self,
        addr: Optional[int] = None,
        interval: int = 0x6,
        sup_timeout: int = 0x64,
        conn_params: Optional[ConnParams] = None,
    ) -> StatusCode:
        """Initialize a connection.

        Convenience function which sends a sequence of commands to
        the DUT, telling it to initialize a connection. PHYs preferences
        cannot be set when using this function, but connection parameters
        can be using the optional `conn_params` parameter. If a value is
        provided for `conn_params`, the values of the `addr`, `interval`,
        and `sup_timeout` parameters are ignored. If no value is provided,
        all connection parameters except min/max interval and supervision
        timeout are defaulted. In addition, a value for `addr` must be
        provided.

        Parameters
        ----------
        addr : int
            Peer device BD address.
        interval : int, optional
            Connection inverval.
        sup_timeout : int, optional
            Supervision timeout.

        Returns
        -------
        StatusCode
            The return status of the create connection command.

        Raises
        ------
        ValueError
            If both `addr` and `conn_params` are None.
        ValueError
            If `addr` is more than 6 bytes in size.

        """
        if conn_params is None:
            if addr is None:
                raise ValueError(
                    "Either connection parameters or address must be provided."
                )

            if max((addr.bit_length() + 7) // 8, 1) > 6:
                raise ValueError(
                    f"Address ({addr}) is too large, must be 6 bytes or less."
                )

            conn_params = ConnParams(
                addr,
                conn_interval_max=interval,
                conn_interval_min=interval,
                sup_timeout=sup_timeout,
            )

        self.reset_connection_stats()
        self.set_default_phy(all_phys=0, tx_phys=7, rx_phys=7)

        status = self.create_connection(conn_params)

        return status

    def read_event(self, timeout: Optional[float] = None) -> EventPacket:
        """Read an event from controller.

        Parameters
        ----------
        timeout : Optional[float], optional
            Timeout for read operation. Can be used to
            temporarily override this object's `timeout`
            attribute.

        Returns
        -------
        EventPacket
            Packet retrieved from the controller.

        Raises
        ------
        TimeoutError
            If a timeout occurs and there are no retries remaining.

        """
        timeout_err = None
        tries = self.retries
        if not timeout:
            timeout = self.timeout
        while tries >= 0:
            try:
                return self.port.retrieve_packet(timeout=self.timeout)
            except TimeoutError as err:
                tries -= 1
                timeout_err = err
                self.logger.warning(
                    "Timeout occured. Retrying. %d retries remaining.",
                    self.retries - tries,
                )

        raise TimeoutError("Timeout occured. No retries remaining.") from timeout_err

    def write_command(
        self,
        command: CommandPacket,
        timeout: Optional[float] = None,
    ) -> EventPacket:
        """Send a command and retrieve the return packet.

        Parameters
        ----------
        command : Union[str, int]
            Command to send. Must be an instance of the
            `CommandPacket` class.
        timeout : int
            Timeout for read portion of the read/write.
            Can be used to temporarily override this object's
            `timeout` attribute.

        Returns
        -------
        EventPacket
            The command return packet.

        """
        if not timeout:
            timeout = self.timeout
        evt = self.port.send_command(command, timeout=timeout)

        return evt

    def write_command_raw(
        self,
        raw_command: bytearray,
        timeout: Optional[float] = None,
    ) -> EventPacket:
        """Write raw command to device

        Parameters
        ----------
        raw_command : bytearray
            Command as bytearray
        timeout : int
            Timeout for read portion of the read/write.
            Can be used to temporarily override this object's
            `timeout` attribute.

        Returns
        -------
        EventPacket

        """

        if not timeout:
            timeout = self.timeout
        return self.port.send_command_raw(raw_command, timeout)

    def exit(self) -> None:
        """Close the HCI connection.

        Used to safely close the connection between the HCI and
        the test board.

        """
        self.port.close()

    def _init_ports(
        self,
        port_id: str,
        baud: int,
        logger_name: str,
        async_callback: Optional[Callable[[AsyncPacket], Any]],
        evt_callback: Optional[Callable[[EventPacket], Any]],
    ) -> None:
        """Initializes serial ports.

        PRIVATE

        """
        self.port = SerialUartTransport(
            port_id,
            baud=baud,
            id_tag=self.id_tag,
            logger_name=logger_name,
            retries=self.retries,
            timeout=self.timeout,
            async_callback=async_callback,
            evt_callback=evt_callback,
        )
